﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Mbox2Eml.Properties;

namespace Mbox2Eml
{
	public partial class MainForm : Form
	{
		private ETACalculator ETACalculator;

		public MainForm()
		{
			InitializeComponent();
		}

		#region Save / Load
		private void MainForm_Load(object sender, EventArgs e)
		{
			this.txtMboxFolder.Text = Settings.Default.MbxFolder;
			this.txtOutputFolder.Text = Settings.Default.OutputFolder;
		}

		private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			Settings.Default.MbxFolder = this.txtMboxFolder.Text;
			Settings.Default.OutputFolder = this.txtOutputFolder.Text;

			Settings.Default.Save();
		}
		#endregion

		private void btnSelectMboxFolder_Click(object sender, EventArgs e)
		{
			if (DialogResult.OK == folderBrowserDialogMbox.ShowDialog())
			{
				this.txtMboxFolder.Text = folderBrowserDialogMbox.SelectedPath;
			}
		}

		private void btnSelectoutputFolder_Click(object sender, EventArgs e)
		{
			if (DialogResult.OK == folderBrowserDialogOutput.ShowDialog())
			{
				this.txtOutputFolder.Text = folderBrowserDialogOutput.SelectedPath;
			}
		}

		private void BeginBlock()
		{
			this.Cursor = Cursors.WaitCursor;
			this.btnSelectMboxFolder.Enabled = false;
			this.btnSelectOutputFolder.Enabled = false;
		}

		private void EndBlock()
		{
			this.Cursor = Cursors.Default;
			this.btnSelectMboxFolder.Enabled = true;
			this.btnSelectOutputFolder.Enabled = true;
		}

		private void btnConvert_Click(object sender, EventArgs e)
		{
			string mboxfolder = this.txtMboxFolder.Text;
			string outputfolder = this.txtOutputFolder.Text;

			if (!string.IsNullOrWhiteSpace(mboxfolder) && !string.IsNullOrWhiteSpace(outputfolder))
			{
				if (mboxfolder == outputfolder)
				{
					MessageBox.Show("'Local Mbox Folder' and 'Output Folder' are the same.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Warning);
					return;
				}

				if (backgroundWorker.IsBusy)
				{
					backgroundWorker.CancelAsync();
				}
				else
				{
					this.BeginBlock();

					this.lblProgressInfo.Text = string.Empty;
					this.lblRemaining.Text = "...";
					this.ETACalculator = new ETACalculator(100, 10);

					progressBar.ForeColor = Color.Green;
					progressBar.Value = 0;

					backgroundWorker.RunWorkerAsync(new { mboxFolder = mboxfolder, outputFolder = outputfolder });
					btnConvert.Text = "Cancel";
				}
			}
			else
			{
				MessageBox.Show("'Local Mbox Folder' or 'Output Folder' is missing.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Warning);
			}
		}

		#region BackGroundWorker
		private void backgroundWorker_DoWork(object sender, DoWorkEventArgs eventArgs)
		{
			dynamic settings = eventArgs.Argument;
			string mboxFolder = settings.mboxFolder;
			string outputFolder = settings.outputFolder;

			try
			{
				var files = MboxUtils.GetMboxFiles(mboxFolder);

				int idx = 0;
				long totalSize = 0;
				foreach (var file in files)
				{
					totalSize += new FileInfo(file).Length;
				}

				long lengthRead = 0;
				foreach (var file in files)
				{
					string directoryName = Path.GetDirectoryName(file);
					string folderName = directoryName != mboxFolder ? directoryName.Substring(mboxFolder.Length + 1).Replace(".sbd", string.Empty) : Path.GetFileName(file);
					string outputPath = folderName != null ? Path.Combine(outputFolder, folderName) : outputFolder;

					var reader = File.OpenText(file);
					string line = null; // reader.ReadLine();
					bool firstLine = true;

					var sb = new StringBuilder();
					while (!reader.EndOfStream)
					{
						if (backgroundWorker.CancellationPending)
						{
							eventArgs.Cancel = true;
							return;
						}

						line = reader.ReadLine();

						if ((line.StartsWith("From - ") || reader.EndOfStream) && !firstLine)
						{
							string text = sb.ToString();
							string messageIdx = "MSG_" + idx.ToString("D8");

							try
							{
								MboxUtils.ProcessMessage(outputPath, messageIdx, text);
							}
							catch
							{
								File.WriteAllText(Path.Combine(outputPath, messageIdx + ".ERROR"), text);
							}

							lengthRead += sb.Length;

							double pct = (100.0 * lengthRead) / totalSize;
							backgroundWorker.ReportProgress((int)pct, new { file = folderName, lengthRead = lengthRead, totalSize = totalSize, idx = idx });

							sb.Clear();
							idx++;
						}
						else
						{
							sb.AppendLine(line);
						}

						firstLine = false;
					}
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(string.Format("There was an error during conversion : {0}", ex.Message),
					"Error", MessageBoxButtons.OK, MessageBoxIcon.Stop);

				eventArgs.Cancel = true;
			}
		}

		private void backgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
		{
			int currentValue = this.progressBar.Value;
			this.progressBar.Value = e.ProgressPercentage;

			dynamic info = e.UserState;
			string file = Path.GetFileName(info.file) ?? Path.GetFileName(info.file);
			long lengthRead = info.lengthRead;
			long totalSize = info.totalSize;
			string idx = info.idx.ToString("D8");

			this.lblProgressInfo.Text = string.Format("Processing file/folder '{0}', message {1}  ({2} from {3}) ({4}%)",
				file,
				idx,
				Utils.StrFormatByteSize(lengthRead),
				Utils.StrFormatByteSize(totalSize),
				e.ProgressPercentage
				);

			this.ETACalculator.Update((float)lengthRead / (float)totalSize);

			if (this.ETACalculator.ETAIsAvailable)
			{
				TimeSpan etr = this.ETACalculator.ETR;
				this.lblRemaining.Text = string.Format("{0:D2} hrs, {1:D2} mins, {2:D2} secs remaining", etr.Hours, etr.Minutes, etr.Seconds);
			}
		}

		private void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
		{
			if (e.Cancelled)
			{
				progressBar.ForeColor = Color.Red;
				progressBar.Value = 0;
			}
			else
			{
				progressBar.ForeColor = Color.Green;
				progressBar.Value = 100;
			}

			this.btnConvert.Enabled = true;
			this.btnConvert.Text = "Convert";

			this.EndBlock();
		}
		#endregion
	}
}